/*
 * Newt-Puzzle
 * Copyright (C) Aude Jaulent 2008, 2009 <elderme@tuxfamily.org>
 * 
 * Newt-Puzzle is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * Newt-Puzzle is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "mesh.h"

#include "scene.h"


Mesh::Mesh(std::string name, bool colorArray, bool normalArray, bool fogCoordArray, bool tangentArray):
	mName(name), mIsInitialized(false), data(NULL), mpUniformNormal(NULL), mpUniformTangent(NULL), mNbEntities(0) {

	if(colorArray) {
		vertexIds[COLOR_ID] = 1;
	} else {
		vertexIds[COLOR_ID] = 0;
	}		
		
	if(normalArray) {
		vertexIds[NORMAL_ID] = 1;
	} else {
		vertexIds[NORMAL_ID] = 0;
	}
		
	if(fogCoordArray) {
		vertexIds[FOG_COORD_ID] = 1;
	} else {
		vertexIds[FOG_COORD_ID] = 0;
	}
		
	if((tangentArray)&&(scene->getGraphicOptions()->isGLSLEnabled())) {
		vertexIds[TANGENT_ID] = 1;
	} else {
		vertexIds[TANGENT_ID] = 0;
	}		
}

Mesh::~Mesh() {
	if(VBOsupported) {
		glDeleteBuffers(6, vertexIds);
		glDeleteBuffers(geometry.nbIndexArrays, primIds);		
	} else {
		if(data!=NULL) {
			deleteVertexArrayData();
		}
	}
	delete mpUniformNormal;
	delete mpUniformTangent;
	
	delete [] primIds;
}


void Mesh::setName(std::string name) {
	mName = name;
}

std::string Mesh::getName() {
	return mName;
}

int Mesh::getNbEntities() {
	return mNbEntities;
}


void Mesh::deleteVertexArrayData() {
	
	if(data->mpVertices!=NULL) {
		delete [] data->mpVertices;
	}
	if(data->mpColors!=NULL) {
		delete [] data->mpColors;
	}
	if(data->mpNormals!=NULL) {
		delete [] data->mpNormals;
	}
	if(data->mpTexCoord!=NULL) {
		delete [] data->mpTexCoord;
	}
	if(data->mpFogCoord!=NULL) {
		delete [] data->mpFogCoord;
	}
	if(data->mpTangents!=NULL) {
		delete [] data->mpTangents;
	}
	
	if(data->mpIndices!=NULL) {
		for(int i=0;i<geometry.nbIndexArrays;i++) {
			delete [] data->mpIndices[i];
		}
		delete [] data->mpIndices;
	}	
}

void Mesh::initialize(int nbTextures) {
	
	if(!mIsInitialized) {
		initFirstParams();
		
		bool texCoordArray = (nbTextures > 0);
		if(VBOsupported) {
			buildVBOs(texCoordArray);
		} else {
			initVertexArray(texCoordArray);
		}
	}
	
	mIsInitialized = true;
}


bool Mesh::isInitialized() {
	return mIsInitialized;
}


void Mesh::buildVBOs(bool texCoordArray) {

	GLuint *currentId = vertexIds;
	
	glGenBuffers(1, currentId);
	glBindBuffer(GL_ARRAY_BUFFER,vertexIds[VERTEX_ID]);
	glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*sizeof(Vector3),0,GL_STATIC_DRAW);
	createVertices((Vector3 *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
	glUnmapBuffer(GL_ARRAY_BUFFER);

	
	currentId++;
	if(vertexIds[COLOR_ID]) {
		glGenBuffers(1, currentId);
		glBindBuffer(GL_ARRAY_BUFFER,vertexIds[COLOR_ID]);
		glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*sizeof(Color),0,GL_STATIC_DRAW);
		createColors((Color *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ARRAY_BUFFER);
	}
		
	currentId++;
	if(vertexIds[NORMAL_ID]) {
		glGenBuffers(1, currentId);	
		glBindBuffer(GL_ARRAY_BUFFER,vertexIds[NORMAL_ID]);
		glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*sizeof(Vector3),0,GL_STATIC_DRAW);
		createNormals((Vector3 *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ARRAY_BUFFER);		
	}
	
	currentId++;
	if(texCoordArray) {
		glGenBuffers(1, currentId);	
		glBindBuffer(GL_ARRAY_BUFFER,vertexIds[TEX_COORD_ID]);
		glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*2*sizeof(GLfloat),0,GL_STATIC_DRAW);
		createTexCoords((GLfloat *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ARRAY_BUFFER);		
	} else {
		vertexIds[TEX_COORD_ID] = 0;
	}
	
	currentId++;
	if(vertexIds[FOG_COORD_ID]) {
		glGenBuffers(1, currentId);	
		glBindBuffer(GL_ARRAY_BUFFER,vertexIds[FOG_COORD_ID]);
		glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*sizeof(GLfloat),0,GL_STATIC_DRAW);
		createFogCoords((GLfloat *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ARRAY_BUFFER);		
	}
	
	currentId++;
	if(vertexIds[TANGENT_ID]) {
		glGenBuffers(1, currentId);	
		glBindBuffer(GL_ARRAY_BUFFER,vertexIds[TANGENT_ID]);
		glBufferData(GL_ARRAY_BUFFER,geometry.nbVertices*sizeof(Vector3),0,GL_STATIC_DRAW);	
		createTangents((Vector3 *)glMapBuffer(GL_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ARRAY_BUFFER);
	} 	
		
	primIds = new GLuint[geometry.nbIndexArrays];
	glGenBuffers(geometry.nbIndexArrays, primIds);		
		
	for(int i=0;i<geometry.nbIndexArrays;i++) {
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER,primIds[i]);
		glBufferData(GL_ELEMENT_ARRAY_BUFFER,geometry.nbIndices[i]*sizeof(GLuint),0,GL_STATIC_DRAW);
		createIndices (i, (GLuint *)glMapBuffer(GL_ELEMENT_ARRAY_BUFFER,GL_READ_WRITE));
		glUnmapBuffer(GL_ELEMENT_ARRAY_BUFFER);	
	}
	
	glBindBuffer(GL_ARRAY_BUFFER, 0);
}

void Mesh::initVertexArray(bool texCoordArray) {
	data = new VertexArrayData;
	
	data->mpVertices = new Vector3[geometry.nbVertices];
	createVertices(data->mpVertices);
	vertexIds[VERTEX_ID] = 1;
	
	if(vertexIds[NORMAL_ID]) {
		data->mpNormals = new Vector3[geometry.nbVertices];
		createNormals (data->mpNormals);
		vertexIds[NORMAL_ID] = 1;
	} else {
		data->mpNormals = NULL;
	}
	
	if(vertexIds[COLOR_ID]) {
		data->mpColors = new Color[geometry.nbVertices];
		createColors (data->mpColors);
		vertexIds[COLOR_ID] = 1;
	} else {
		data->mpColors = NULL;
	}
	
	if(texCoordArray) {
		data->mpTexCoord = new GLfloat[2*geometry.nbVertices];
		createTexCoords (data->mpTexCoord);
		vertexIds[TEX_COORD_ID] = 1;
	} else {
		data->mpTexCoord = NULL;	
		vertexIds[TEX_COORD_ID] = 0;
	}

	if(vertexIds[FOG_COORD_ID]) {
		data->mpFogCoord = new GLfloat[geometry.nbVertices];
		createFogCoords (data->mpFogCoord);
		vertexIds[FOG_COORD_ID] = 1;
	} else {
		data->mpFogCoord = NULL;
	}	
	
	if(vertexIds[TANGENT_ID]) {
		data->mpTangents = new Vector3[geometry.nbVertices];
		createTangents (data->mpTangents);
		vertexIds[TANGENT_ID] = 1;
	} else {
		data->mpTangents = NULL;
	}
	
	data->mpIndices = new GLuint * [geometry.nbIndexArrays];
	primIds = new GLuint[geometry.nbIndexArrays];
		
	for(int i=0;i<geometry.nbIndexArrays;i++) {
		data->mpIndices[i] = new GLuint[geometry.nbIndices[i]];
	}
		
	for(int i=0;i<geometry.nbIndexArrays;i++) {
		createIndices (i, data->mpIndices[i]);
		primIds[i] = 1;
	}	
		
}


void Mesh::activateClientStates() {
	glEnableClientState(GL_VERTEX_ARRAY);
	if(vertexIds[NORMAL_ID]) {		
		glEnableClientState(GL_NORMAL_ARRAY);
	}
	if(vertexIds[COLOR_ID]) {
		glEnableClientState (GL_COLOR_ARRAY);
	}
	if(vertexIds[FOG_COORD_ID]) {
		glEnableClientState(GL_FOG_COORD_ARRAY);
	}
	if((shaderMana->getCurrentShaderType()!=FIXED_PIPELINE)&&(vertexIds[TANGENT_ID])) {
		glEnableVertexAttribArray(shaderMana->getLocationTangent());
	}
}


void Mesh::draw(int nbTextures) {
	
	glPushClientAttrib (GL_CLIENT_VERTEX_ARRAY_BIT);
	activateClientStates();
	
	if(VBOsupported) {
		drawWithVBO(nbTextures);
	} else {
		drawWithVertexArrays(nbTextures);
	}
		
	glPopClientAttrib ();
}

void Mesh::drawWithVBO(int nbTextures) {
	
	if(vertexIds[NORMAL_ID]!=0) {			
		glBindBuffer(GL_ARRAY_BUFFER, vertexIds[NORMAL_ID]);
		glNormalPointer(GL_FLOAT, sizeof(Vector3), NULL);
	} else {
		if(mpUniformNormal!=NULL) {
			glNormal3f(mpUniformNormal->x, mpUniformNormal->y, mpUniformNormal->z);
		} 
	}
	
	if((shaderMana->getCurrentShaderType()!=FIXED_PIPELINE)) {
		if(vertexIds[TANGENT_ID] !=0) {
			glBindBuffer(GL_ARRAY_BUFFER, vertexIds[TANGENT_ID]);
			glVertexAttribPointer(shaderMana->getLocationTangent(), 3, GL_FLOAT, GL_FALSE, sizeof(Vector3), NULL);
		} else {	
			if(mpUniformTangent!=NULL) {
				glVertexAttrib3f(shaderMana->getLocationTangent(),mpUniformTangent->x, mpUniformTangent->y, mpUniformTangent->z);
			}
		}
	}
	
	if(nbTextures != 0) {
		glBindBuffer(GL_ARRAY_BUFFER, vertexIds[TEX_COORD_ID]);
		for(int i=0;i<=nbTextures;i++) {
			glClientActiveTexture(GL_TEXTURE0 + i);
			glTexCoordPointer(2, GL_FLOAT, 2*sizeof(GLfloat), NULL);
			glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		}
	}
	
	//TODO the other parameters...
	

	glBindBuffer(GL_ARRAY_BUFFER, vertexIds[VERTEX_ID]);
	glVertexPointer(3, GL_FLOAT, sizeof(Vector3), NULL);
	
	for(int i=0;i<geometry.nbIndexArrays;i++) {		
		glBindBuffer(GL_ELEMENT_ARRAY_BUFFER, primIds[i]);
		glDrawElements(geometry.primType[i], geometry.nbIndices[i], GL_UNSIGNED_INT, NULL);
	}	
	
}

void Mesh::drawWithVertexArrays(int nbTextures) {
	
	glVertexPointer(3, GL_FLOAT, sizeof(Vector3), data->mpVertices);
	
	if(vertexIds[NORMAL_ID]) {
		glNormalPointer(GL_FLOAT, sizeof(Vector3), data->mpNormals);
	}	
	 else {
		 if(mpUniformNormal != NULL) {
			glNormal3f(mpUniformNormal->x, mpUniformNormal->y, mpUniformNormal->z);
		} 
	}
	
	if((shaderMana->getCurrentShaderType()!=FIXED_PIPELINE)) {
		if(vertexIds[TANGENT_ID]) {
			//glVertexAttribPointer(shaderMana->getLocationTangent(), 3, GL_FLOAT, GL_FALSE, sizeof(Vector3), NULL);
			glVertexAttribPointer(shaderMana->getLocationTangent(), 3, GL_FLOAT, GL_FALSE, sizeof(Vector3), data->mpTangents);

		} else {
		
			if(mpUniformTangent != NULL) {
				glVertexAttrib3f(shaderMana->getLocationTangent(),mpUniformTangent->x, mpUniformTangent->y, mpUniformTangent->z);
			} 
		}
	}
	
	
	if(vertexIds[COLOR_ID]) {
		glColorPointer (4, GL_FLOAT, sizeof(Color), data->mpColors);
	}
	
	/*if(vertexIds[FOG_COORD_ID) {
		// TODO
	}*/
	
	for(int i=0;i<nbTextures;i++) {
		glClientActiveTexture(GL_TEXTURE0 + i);
		glEnableClientState(GL_TEXTURE_COORD_ARRAY);
		glTexCoordPointer (2, GL_FLOAT, 2*sizeof(float), data->mpTexCoord);
	}

	for(int i=0;i<geometry.nbIndexArrays;i++) {	
		glDrawElements(geometry.primType[i], geometry.nbIndices[i], GL_UNSIGNED_INT, data->mpIndices[i]);
	}	
}


void Mesh::incNbEntities() {
	mNbEntities++;
}

void Mesh::decNbEntities() {
	mNbEntities--;
}
